/*
Copyright (c) 2014, Romain Dolbeau, unless otherwise noted.
No claims is made upon work by others.

For the work by Romain Dolbeau:
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the work by others:
See the indicated reference for the relevant license.
*/

#include <stdint.h>

#include "epi_crypto_common.h"

#include "e_epi_crypto.h"

#include "chacha20_impl.h"

#ifdef __HOST__
#include <e-hal.h>
#include <string.h>
#include <openssl/evp.h>
#else
#include <e_lib.h>
#endif

/* chacha20 implementation based on code by D. J. Bernstein */

#define U32C(v) (v##U)
#define U32TO32_LITTLE(v) (v)
#define U8TO32_LITTLE(p) U32TO32_LITTLE(((uint32_t*)(p))[0])
#define U32TO8_LITTLE(p, v) (((uint32_t*)(p))[0] = U32TO32_LITTLE(v))

#define U32V(v) ((uint32_t)(v) & U32C(0xFFFFFFFF))
#define ROTL32(v, n) \
  (U32V((v) << (n)) | ((v) >> (32 - (n))))

#ifdef __HOST__
#define ROTATE(v,c) (ROTL32(v,c))
#define XOR(v,w) ((v) ^ (w))
#define PLUS(v,w) (U32V((v) + (w)))
#define PLUSONE(v) (PLUS((v),1))
#else
//#define NO_IALU2
#ifdef NO_IALU2
#define ROTATE(v,c) (ROTL32(v,c))
#define PLUS(v,w) (U32V((v) + (w)))
#else
static inline uint32_t ROTATE(const uint32_t v, const uint32_t c) {
  const uint32_t twoc = 1 << c;
  uint32_t r;
  r = v >> (32-c);
  __asm__("IMADD %0,%1,%2" : "+r"(r) : "r"(v), "r"(twoc));
  return r;
}
static inline uint32_t PLUS(const uint32_t v, const uint32_t w) {
  uint32_t r;
  __asm__("IADD %0,%1,%2" : "=r"(r) : "r"(v), "r"(w));
  return r;
}
#endif
#define PLUSONE(v) (PLUS((v),1))
#define XOR(v,w) ((v) ^ (w))
#endif

#define QUARTERROUND(a,b,c,d) \
  x[a] = PLUS(x[a],x[b]); x[d] = ROTATE(XOR(x[d],x[a]),16); \
  x[c] = PLUS(x[c],x[d]); x[b] = ROTATE(XOR(x[b],x[c]),12); \
  x[a] = PLUS(x[a],x[b]); x[d] = ROTATE(XOR(x[d],x[a]), 8); \
  x[c] = PLUS(x[c],x[d]); x[b] = ROTATE(XOR(x[b],x[c]), 7);

static inline void salsa20_wordtobyte(uint8_t output[64],const uint32_t input[16])
{
  uint32_t x[16];
  int i;

  for (i = 0;i < 16;++i) x[i] = input[i];
  for (i = 20;i > 0;i -= 2) {
    QUARTERROUND( 0, 4, 8,12)
    QUARTERROUND( 1, 5, 9,13)
    QUARTERROUND( 2, 6,10,14)
    QUARTERROUND( 3, 7,11,15)
    QUARTERROUND( 0, 5,10,15)
    QUARTERROUND( 1, 6,11,12)
    QUARTERROUND( 2, 7, 8,13)
    QUARTERROUND( 3, 4, 9,14)
  }
  for (i = 0;i < 16;++i) x[i] = PLUS(x[i],input[i]);
  for (i = 0;i < 16;++i) U32TO8_LITTLE(output + 4 * i,x[i]);
}

#ifdef __HOST__
#ifdef __ARM_NEON__
#include <arm_neon.h>
#endif
static inline void encrypt_bytes(uint32_t *x, const uint8_t *m, uint8_t *c_, uint32_t bytes)
{
  uint8_t output[64];
  int i;
  uint8_t* out = c_;

#ifdef __ARM_NEON__
#if 1
#include "u4_neon.h"
#endif
#if 0
#include "u1_neon.h"
#endif
#endif

#ifdef __x86_64__
#if defined(__AVX512F__)
#include "u16_x86_64.h"
#endif
#if defined(__AVX2__)
#include "u8_x86_64.h"
#endif
#if 1
#include "u4_x86_64.h"
#endif
#if 1
#include "u1_x86_64.h"
#endif
#endif

#ifdef __MIPSEL__
#if defined(__mips_msa)
#include "u4_msa.h"
#endif
#endif

  if (!bytes) return;
  for (;;) {
    salsa20_wordtobyte(output,x);
    x[12] = PLUSONE(x[12]);
    if (!x[12]) {
      x[13] = PLUSONE(x[13]);
      /* stopping at 2^70 bytes per nonce is user's responsibility */
    }
    if (bytes <= 64) {
      for (i = 0;i < bytes;++i) out[i] = m[i] ^ output[i];
      return;
    }
    for (i = 0;i < 64;++i) out[i] = m[i] ^ output[i];
    bytes -= 64;
    out += 64;
    m += 64;
  }
}

static const char sigma[16] = "expand 32-byte k";
static const char tau[16] = "expand 16-byte k";
static inline void keysetup(uint32_t *input, const uint8_t *k, const uint32_t kbits, const uint32_t ivbits)
{
  const char *constants;

  input[4] = U8TO32_LITTLE(k + 0);
  input[5] = U8TO32_LITTLE(k + 4);
  input[6] = U8TO32_LITTLE(k + 8);
  input[7] = U8TO32_LITTLE(k + 12);
  if (kbits == 256) { /* recommended */
    k += 16;
    constants = sigma;
  } else { /* kbits == 128 */
    constants = tau;
  }
  input[8] = U8TO32_LITTLE(k + 0);
  input[9] = U8TO32_LITTLE(k + 4);
  input[10] = U8TO32_LITTLE(k + 8);
  input[11] = U8TO32_LITTLE(k + 12);
  input[0] = U8TO32_LITTLE(constants + 0);
  input[1] = U8TO32_LITTLE(constants + 4);
  input[2] = U8TO32_LITTLE(constants + 8);
  input[3] = U8TO32_LITTLE(constants + 12);
}

static inline void ivsetup(uint32_t *input, const uint8_t *iv)
{
  input[12] = 0;
  input[13] = 0;
  input[14] = U8TO32_LITTLE(iv + 0);
  input[15] = U8TO32_LITTLE(iv + 4);
}

void chacha20_prepare_epi_keys(const uint32_t key[], const uint32_t IV[], my_epiphany* epip) {
  uint32_t cstate[16];
  size_t r;
  keysetup(cstate, key, 256, 0);
  ivsetup(cstate, IV);

  E_CHECK_ERR(e_write(&epip->DRAM, 0, 0, offsetof(shared_buf_t, k), (void *)cstate, sizeof(cstate)));  
}

void chacha20_host_encrypt(const uint32_t key[], const uint32_t IV[], const uint8_t* in, uint8_t *out, size_t size) {
  uint32_t cstate[16];
  keysetup(cstate, key, 256, 0);
  ivsetup(cstate, IV);

  encrypt_bytes(cstate, in, out, size);
}
#endif

#ifndef __HOST__

int main(int argc, char **argv) {

  init();

  while (1) {
    while (Mailbox.pCore->cn[me.corenum] == 0) {
      /* wait */
    }
    uint64_t cs = Mailbox.pCore->cs[me.corenum];
    uint64_t cn = Mailbox.pCore->cn[me.corenum];
    //uint64_t cd = Mailbox.pCore->cd[me.corenum];
    uint64_t mcd = 0;

#define ISOLATE_(X) \
    for (int i = 0 ; i < 16 ; i++) {            \
      e_barrier(barriers, tgt_bars);            \
      if (me.corenum == i)                      \
        X;                                      \
    }
#define ISOLATE(X) X;
    
    ISOLATE(e_dma_copy(lk, Mailbox.k, 64))
    
    ISOLATE(Mailbox.pCore->cn[me.corenum] = 0)

#ifndef NO_IALU2
    setIntegerMode(1);
#endif

    e_irq_attach(E_USER_INT, empty_isr);
    
    for (uint64_t l = 0 ; l < cn ; l += PCSSIZE/64) {
      uint64_t scount = 0;
      for (uint64_t j = 0 ; j < PCSSIZE/64 && (j+l) < cn ; j += BSIZE/64) {
        uint64_t count = 0;
        for (uint64_t k = 0 ; k < BSIZE/64 && (k+j+l) < cn; k++) {
          uint64_t i = l + j + k;
          uint64_t ctr = cs + i;
          lk[13] = (uint32_t)(0xFFFFFFFF & (ctr>>32));
          lk[12] = (uint32_t)(0xFFFFFFFF & (ctr    ));
          salsa20_wordtobyte(lc+(16*k), lk);
          count++;
        }
        ISOLATE(e_dma_copy(Mailbox.c+(me.corenum*PCSSIZE/sizeof(uint32_t))+(scount*16), lc, 64*count))
        mcd += count;
        scount += count;
        ISOLATE(Mailbox.pCore->cd[me.corenum] = mcd)
      }
      wait_on_user();
    }
  }
  
  return 0;
}
#endif // __HOST__
