/*
Copyright (c) 2014, Romain Dolbeau, unless otherwise noted.
No claims is made upon work by others.

For the work by Romain Dolbeau:
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the work by others:
See the indicated reference for the relevant license.
*/

#define _BSD_SOURCE
#include <stdlib.h>
#undef _BSD_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stddef.h>
#include <malloc.h>
#include <unistd.h>

#include <e-hal.h>
#include <e-loader.h>

#include <sys/time.h>
#include <time.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <arm_neon.h>

#include "epi_crypto_common.h"

#include "aes256_impl.h"
#include "chacha20_impl.h"

#define MMAP_SHARED


static inline double wallclock(void) {
  struct timeval tv;
  struct timezone tz;
  double t;
  static double tr;
  
  gettimeofday(&tv, &tz);
  if (tr == 0.)
    tr = (double) tv.tv_sec;
    
  t = (double) tv.tv_sec - (double)tr;
  t += ((double) tv.tv_usec) / 1000000.0;
  
  return t;
}

typedef struct  {
  const char *name;
  prepare_epi_keys prepare_epi;
  host_encrypt encrypt;
  host_encrypt ref;
  uint32_t bs;
} algo;

const algo algos[] = {
  {"aes256"  , &aes256_prepare_epi_keys  , &aes256_host_encrypt  , &aes256_openssl_encrypt, 16 },
  {"chacha20", &chacha20_prepare_epi_keys, &chacha20_host_encrypt, NULL                   , 64 },
  {NULL      , NULL                      , NULL                  , NULL                   , 0 }
};

static inline void mymemcpy_m16(char *dst, const char *src, const size_t size) {
  size_t i = 0;
#if 0
  for ( ; i < size-112 ; i+= 128) {
    uint32x4x2_t t0 = vld2q_u32((const uint32_t*)(src+i+ 0));
    uint32x4x2_t t1 = vld2q_u32((const uint32_t*)(src+i+32));
    uint32x4x2_t t2 = vld2q_u32((const uint32_t*)(src+i+64));
    uint32x4x2_t t3 = vld2q_u32((const uint32_t*)(src+i+96));
    vst2q_u32((uint32_t*)(dst+i+ 0), t0);
    vst2q_u32((uint32_t*)(dst+i+32), t1);
    vst2q_u32((uint32_t*)(dst+i+64), t2);
    vst2q_u32((uint32_t*)(dst+i+96), t3);
  }
#endif
#if 0
   for ( ; i < size-127 ; i+= 128) {
    uint32x4x4_t t0 = vld4q_u32((const uint32_t*)(src+i+ 0));
    uint32x4x4_t t1 = vld4q_u32((const uint32_t*)(src+i+64));
    vst4q_u32((uint32_t*)(dst+i+ 0), t0);
    vst4q_u32((uint32_t*)(dst+i+64), t1);
  }

#endif
#if 1
  for ( ; i < size-63 ; i+= 64) {
    uint32x4x2_t t0 = vld2q_u32((const uint32_t*)(src+i+ 0));
    uint32x4x2_t t1 = vld2q_u32((const uint32_t*)(src+i+32));
    vst2q_u32((uint32_t*)(dst+i+ 0), t0);
    vst2q_u32((uint32_t*)(dst+i+32), t1);
  }
#else
  for ( ; i < size-63 ; i+= 64) {
    uint32x4_t t0 = vld1q_u32((const uint32_t*)(src+i+ 0));
    uint32x4_t t1 = vld1q_u32((const uint32_t*)(src+i+16));
    uint32x4_t t2 = vld1q_u32((const uint32_t*)(src+i+32));
    uint32x4_t t3 = vld1q_u32((const uint32_t*)(src+i+48));
    vst1q_u32((uint32_t*)(dst+i+ 0), t0);
    vst1q_u32((uint32_t*)(dst+i+16), t1);
    vst1q_u32((uint32_t*)(dst+i+32), t2);
    vst1q_u32((uint32_t*)(dst+i+48), t3);
  }
#endif
  for ( ; i < size ; i+= 16) {
    uint32x4_t t0 = vld1q_u32((const uint32_t*)(src+i+ 0));
    vst1q_u32((uint32_t*)(dst+i+ 0), t0);
  }
}

static inline void mymemcpy_xor(char *dst, const char *src1, const char* src2, const size_t size) {
  size_t i = 0;
  for ( ; i < size-31 ; i+= 32) {
    uint32x4_t t0a = vld1q_u32((const uint32_t*)(src1+i+ 0));
    uint32x4_t t0b = vld1q_u32((const uint32_t*)(src2+i+ 0));
    uint32x4_t t1a = vld1q_u32((const uint32_t*)(src1+i+16));
    uint32x4_t t1b = vld1q_u32((const uint32_t*)(src2+i+16));
    uint32x4_t t0c, t1c;
    
    t0c = veorq_u32(t0a, t0b);
    t0c = veorq_u32(t0a, t0b);
    t1c = veorq_u32(t1a, t1b);
    t1c = veorq_u32(t1a, t1b);
    
    vst1q_u32((const uint32_t*)(dst+i+ 0), t0c);
    vst1q_u32((const uint32_t*)(dst+i+16), t1c);
  }
  for ( ; i < size ; i++) {
    dst[i] = src1[i] ^ src2[i];
  }
}
#define memcpy(a,b,c) mymemcpy_m16(a,b,c)
/* assume key are set in e */
void epi_do_encrypt(my_epiphany* epip,
                    const uint32_t bs, /* algo block size */
                    const size_t size,
                    const uint8_t* in,
                    uint8_t* out) {
  uint64_t blocks = (size+bs-1)/bs;
  uint64_t eblocks = ((blocks+NCORE-1)/NCORE)*NCORE;
  uint64_t pcbl = eblocks/NCORE;
  uint64_t cs[NCORE], cn[NCORE], cd[NCORE], ocd[NCORE]; /* start, number, done, olddone */
  size_t j;
  uint32_t addr;
  vuint32_t ready;
  size_t r;
  int kg = 1;
  int i;
  for (i = 0 ; i < NCORE ; i++) {
    cs[i] = i * pcbl;
    cn[i] = pcbl;
    cd[i] = 0;
    ocd[i] = 0;
  }
  addr = offsetof(shared_buf_t, core.ready);
  ready = 0;
  while (ready == 0) {
    E_CHECK_ERR(e_read(&epip->DRAM, 0, 0, addr, (void *)(&ready), sizeof(ready)));
  }

  addr = offsetof(shared_buf_t, core.cs); /* set start */
  E_CHECK_ERR(e_write(&epip->DRAM, 0, 0, addr, (void *)cs, sizeof(uint64_t)*NCORE));
  addr = offsetof(shared_buf_t, core.cn); /* set number - epiphany starts crunching */
  E_CHECK_ERR(e_write(&epip->DRAM, 0, 0, addr, (void *)cn, sizeof(uint64_t)*NCORE));

/*   if (pcbl*bs>PCSSIZE) */
/*     usleep(PCSSIZE/20); */
/*   else */
/*     usleep(pcbl*bs/20); */

  while (kg) {
    addr = offsetof(shared_buf_t, core.cd);
    /* check status for everyone */
    E_CHECK_ERR(e_read(&epip->DRAM, 0, 0, addr, (void *)cd, sizeof(uint64_t)*NCORE));
    for (i = 0 ; i < NCORE ; i++) {
      if (cd[i] > ocd[i]) { /* we have new data */
        if (cd[i] > (size/bs)) {
          fprintf(stderr, "%llu (0x%016llx) is not an acceptable value from %d...\n", cd[i], cd[i], i);
        } else {
          uint64_t nbytes = bs*(cd[i]-ocd[i]);
          if (nbytes > PCSSIZE) {
            fprintf(stderr, "Going straight from %llu to %llu (%llu bytes) is not legit for %d...\n", ocd[i], cd[i], nbytes, i);
            nbytes = PCSSIZE;
            cd[i] = ocd[i]+(PCSSIZE/bs);
          }
          uint64_t rnbytes = nbytes;
          size_t offset = cs[i]*bs+ocd[i]*bs;
          size_t last = rnbytes + offset;
          if (last > size) {
            size_t diff = last - size;
            if (diff > rnbytes)
              rnbytes = 0;
            else
              rnbytes -= diff;
          }
          if (rnbytes > 0) {
            addr = offsetof(shared_buf_t, c) + i * PCSSIZE; // beginning of core zone
            addr += ((ocd[i]*bs) % PCSSIZE); // offset in core zone
#ifdef MMAP_SHARED
            if (in != NULL) {
#if 0
              memcpy((out+offset), ((uint8_t*)epip->ptr)+addr, nbytes); // full block
              for (j = 0 ; j < rnbytes ; j++) {
                /* directly addressing the shared memory via byte addressing
                   is *very slow...
                   need to work a better way... */
                /*               out[j+offset] = (((uint8_t*)epip->ptr)+addr)[j] ^ in[j+offset]; */
                out[j+offset] ^= in[j+offset];
              }
#else
              mymemcpy_xor(out+offset, ((uint8_t*)epip->ptr)+addr, in+offset, rnbytes);
#endif
            } else {
              memcpy((out+offset), ((uint8_t*)epip->ptr)+addr, rnbytes);
            }
#else
            E_CHECK_ERR(e_read(&epip->DRAM, 0, 0, addr, (void *)(out+offset), nbytes)); // full block
#endif
          }
#ifdef SYNCHRO_BY_SIGNAL
            if (((ocd[i]*bs+nbytes)%PCSSIZE == 0) || (ocd[i] >= cn[i])) {
              unsigned row, col;
              e_get_coords_from_num(dev, i, &row, &col);
              E_CHECK_ERR(e_signal(dev, row, col));
              E_CHECK_ERR(e_signal(dev, row, col)); /* fixme */
            }
#else
          if (((ocd[i]*bs+nbytes)%PCSSIZE == 0) || (ocd[i] >= cn[i])) {
            uint64_t zero = 0;
            addr = offsetof(shared_buf_t, core.cd) + i*sizeof(uint64_t);
            E_CHECK_ERR(e_write(&epip->DRAM, 0, 0, addr, (void *)&zero, sizeof(uint64_t)));
          }
#endif
#ifdef MMAP_SHARED
          /* done above */
#else
          /* do XORing with the input message */
          if (in != NULL) {
            for (j = 0 ; j < rnbytes ; j++) {
              out[j+offset] ^= in[j+offset];
            }
          }
#endif
          ocd[i] = cd[i];
          cd[i] = 0;
        }
      }
    }
    kg = 0;
    for (i = 0 ; i < NCORE && !kg ; i++) {
      kg = ocd[i] < cn[i];
    }
  }
}
#undef memcpy

void epi_init(my_epiphany* epip) {
  vuint32_t ready;
  size_t r;
  e_set_host_verbosity(H_D0);
  e_init(NULL);
  e_reset_system();
  e_get_platform_info(&epip->platform);

  if (e_open(&epip->dev, 0, 0, epip->platform.rows, epip->platform.cols)) {
    fprintf(stderr, "ERROR: Can't establish connection to Epiphany device!\n");
    exit(-1);
  }
  if (e_alloc(&epip->DRAM, 0x00000000, sizeof(shared_buf_t))) {
    fprintf(stderr, "ERROR: Can't allocate Epiphany DRAM!\n");
    exit(-2);
  }

  ready = 0;
  E_CHECK_ERR(e_write(&epip->DRAM, 0, 0, offsetof(shared_buf_t, core.ready), (void *)(&ready), sizeof(ready)));

#ifdef MMAP_SHARED
  {
    unsigned int shared_addr = 0x3e000000; // fixme: BEWARE ! direct physical addressing...
    off_t page_addr;
    size_t page_size=sysconf(_SC_PAGESIZE);
/*     size_t ssize = 32*1024*1024; */
    size_t ssize = (sizeof(shared_buf_t)+page_size-1) & (~(page_size-1));
    epip->fd = open ("/dev/mem", O_RDONLY);
    page_addr = (shared_addr & (~(page_size-1)));
    epip->page_offset = shared_addr - page_addr;
    epip->ptr = mmap(NULL, ssize, PROT_READ, MAP_SHARED, epip->fd, page_addr);
/*     fprintf(stderr, "mapped %p to %p for size %zd in %d, page_size is %zd\n", */
/*             shared_addr, epip->ptr, ssize, epip->fd, page_size); */
#if 0
    for (size_t offset = 0x01000000 - 4 ; offset <  0x01000000 + 8 ; offset+=4) {
      vuint32_t* ptr = ((vuint32_t*)(((vuint8_t*)epip->ptr)+offset));
      fprintf(stderr, "I'm now going to try to read offset %zd @ %p (%p)...\n", offset, ptr, 0x8e000000+offset);
      vuint32_t x;
      x = *ptr;
      fprintf(stderr, "Value is 0x%08x\n", x);fflush(stderr);
    }
#endif
  }
#endif
}
void epi_stop(my_epiphany* epip) {
  size_t r;

  
  if (e_close(&epip->dev)) {
    fprintf(stderr, "ERROR: Can't close connection to Epiphany device!\n");
    exit(-1);
  }
  if (e_free(&epip->DRAM)) {
    fprintf(stderr, "ERROR: Can't release Epiphany DRAM!\n");
    exit(-2);
  }
#ifdef MMAP_SHARED
  {
    unsigned int page_size=sysconf(_SC_PAGESIZE);
    munmap(epip->ptr, page_size);
    close(epip->fd);
  }
#endif
}

void epi_load_algo(const algo* alg, my_epiphany* epip) {
  char temp[128];
  size_t r;
  snprintf(temp, 128, "e_%s.elf", alg->name);
  
  E_CHECK_ERR(e_load_group(temp, &epip->dev, 0, 0, epip->platform.rows, epip->platform.cols, E_TRUE));
}

void data_compare(const char* name, const char *ha, const char* hb, const uint32_t *a, const uint32_t* b, const size_t size) {
  int header = 0;
  int count = 0;
  for (size_t i = 0 ; i < size/sizeof(uint32_t) && count < 20 ; i++) {
    if (a[i] != b[i]) {
      if (!header) {
        fprintf(stderr, "%8s: % 10s     % 10s\n", name, ha, hb);
        header = 1;
      }
      fprintf(stderr, "@ % 6zd: 0x%08x <-> 0x%08x ", i*sizeof(uint32_t), a[i], b[i]);
      int found = 0;
      for (size_t j = 0 ; j < size/sizeof(uint32_t) && !found ; j++) {
        if (a[j] == b[i]) {
          fprintf(stderr, " @ %zd", j*4);
          found = 1;
        }
      }
      fprintf(stderr, "\n");
      count ++;
    }
  }
}

#define MAX_SIZE (16*1024*1024)
int main(int argc, char **argv) {
  my_epiphany epip;
  const algo* alg;
  uint8_t *in, *oe, *oh, *or;
  uint32_t key[8];
  uint32_t IV[4];
  double t0, t1, t2;
  
  in = (uint8_t*)malloc(MAX_SIZE);
  oe = (uint8_t*)malloc(MAX_SIZE);
  oh = (uint8_t*)malloc(MAX_SIZE);
  or = (uint8_t*)malloc(MAX_SIZE);

  srandom(1);
  for (size_t i = 0 ; i < 8 ; i++) {
    key[i] = random();
  }
  for (size_t i = 0 ; i < 4 ; i++) {
    IV[i] = 0;//random();
  }

  setenv("EPIPHANY_HDF", "/opt/adapteva/esdk/bsps/current/platform.hdf", 0);

  for (int a = 0 ; algos[a].encrypt != NULL ; a++) {
    alg = &algos[a];
  for (size_t size = 2*1024*1024 ; size <= MAX_SIZE ; size *= 2) {
/*   for (size_t size = 13700 ; size <= 500000 ; size *= 7) { */
    for (int init = 0 ; init < 2 ; init ++) {
    printf("Checking '%s' for %zd Bytes, %s\n", alg->name, size, init ? "" : "without xor after Epiphany");
    fflush(stdout);
    srandom(2);
    for (size_t i = 0 ; i < size/4 ; i++) {
      ((uint32_t*)in)[i] = init ? random() : 0;
      ((uint32_t*)oe)[i] = 0xDEADBEEF;
      ((uint32_t*)oh)[i] = 0xDEADBEEF;
      ((uint32_t*)or)[i] = 0xDEADBEEF;
    }
  
    /*   fprintf(stderr, "epi_init()\n"); */
    epi_init(&epip);

    /*   fprintf(stderr, "epi_load_algo()\n"); */
    epi_load_algo(alg, &epip);

    t0 = wallclock();
    /*   fprintf(stderr, "alg->encrypt()\n"); */
    alg->encrypt(key, IV, in, oh, size);
    t1 = wallclock();
    printf("Host impl. : %1.5lf s / %5.1lf MB/s\n", t1-t0, ((double)size)/(1000000.*(t1-t0)));

    t0 = wallclock();
    /*   fprintf(stderr, "alg->prepare()\n"); */
    alg->prepare_epi(key, IV, &epip);
    /*   fprintf(stderr, "epi_do_encrypt()\n"); */
    epi_do_encrypt(&epip, alg->bs, size, init ? in : NULL, oe);
    t1 = wallclock();
    printf("Epip impl. : %1.5lf s / %5.1lf MB/s\n", t1-t0, ((double)size)/(1000000.*(t1-t0)));
  
    /*   fprintf(stderr, "epi_stop()\n"); */
    epi_stop(&epip);

    data_compare(alg->name, "Host", "Epip", (const uint32_t*)oh, (const uint32_t*)oe, size);

    if (alg->ref != NULL) {
      t0 = wallclock();
      /*   fprintf(stderr, "alg->ref()\n"); */
      alg->ref(key, IV, in, or, size);
      t1 = wallclock();
      printf("Ref  impl. : %1.5lf s / %5.1lf MB/s\n", t1-t0, ((double)size)/(1000000.*(t1-t0)));
      data_compare(alg->name, "Ref", "Epip", (const uint32_t*)or, (const uint32_t*)oe, size);
    }
  }
  }
  }

  free(in);
  free(oe);
  free(oh);
  free(or);

  return 0;
}
